from pysondb import db
from tmt.storage.schema import Entry
from typing import List, Optional, Union, Iterable, Generator
from datetime import datetime
import os
import json
import warnings
import re


class DbManager:
    """
    Class to interact with the underlying json database. If `read_only` is `True`, all operations
    which add and/or modify the database won't be allowed.

    :param db_path: path to the json db file .
    :type db_path: str
    :param read_only: if `True` all writing access is denied, defaults to False.
    :type read_only: bool, optional
    """

    def __init__(self, db_path: str, read_only=False):

        os.makedirs(os.path.dirname(db_path), exist_ok=True)
        self.db = db.getDb(db_path)
        self.db._id_fieldname = '__json_id'
        self.read_only = read_only

    def check_can_write(func):
        def inner(*args, **kwargs):
            if not args[0].read_only:
                return func(*args, **kwargs)
        return inner

    @check_can_write
    def add_new_entries(self, entries: List[Entry]):
        self.db.addMany([e.to_dict() for e in entries])
    
    @check_can_write
    def delete_entry(self, entry: Entry) -> bool:
        with self.db.lock:
            with open(self.db.filename, 'r+', encoding='utf-8') as f:
                db_data = json.load(f)
                for i, d in enumerate(db_data['data']):
                    if d['id'] == entry.id:
                        del db_data['data'][i]
                        f.seek(0)
                        f.truncate()
                        json.dump(db_data, f)
                        return True
        return False

    def get_entry_by_id(self, id: str) -> Optional[Entry]:
        if (e := self.db.getByQuery({"id": id})): 
            return Entry.from_dict(e[0])

    def get_entry_by_exact_name(self, name: str) -> Optional[Entry]:
        if (e := self.db.getByQuery({'name': name})):
            if len(e) > 1:
                warnings.warn(f'Found {len(e)} entries for name {name}. Returning the first one.')
            return Entry.from_dict(e[0])

    def get_entries_by_name(self, name: str) -> List[Entry]:
        with self.db.lock:
            with open(self.db.filename, 'r', encoding='utf-8') as f:
                db_data = json.load(f)
            return list(map(Entry.from_dict, filter(lambda d: name in d['name'], db_data['data'])))

    def get_entries_by_name_regex(self, regex: str) -> List[Entry]:
        with self.db.lock:
            with open(self.db.filename, 'r', encoding='utf-8') as f:
                db_data = json.load(f)
            pattern = re.compile(regex)
            return list(map(Entry.from_dict, filter(lambda d: pattern.match(d['name']), db_data['data'])))

    def get_entries_between_dates(self, first: Union[datetime, int], second: Union[datetime, int]) -> List[Entry]:
        with self.db.lock:
            first, second = list(self.__convert_date_to_timestamp(first, second))
            with open(self.db.filename, 'r', encoding='utf-8') as f:
                db_data = json.load(f)
            return list(map(Entry.from_dict, filter(lambda d: first <= d['timestamp'] <= second, db_data['data'])))

    def get_entries_greater_than_date(self, date: Union[datetime, int]) -> List[Entry]:
        with self.db.lock:
            timestamp = next(self.__convert_date_to_timestamp(date))
            with open(self.db.filename, 'r', encoding='utf-8') as f:
                db_data = json.load(f)
            return list(map(Entry.from_dict, filter(lambda d: d['timestamp'] > timestamp, db_data['data'])))

    def get_entries_lower_than_date(self, date: Union[datetime, int]) -> List[Entry]:
        with self.db.lock:
            timestamp = next(self.__convert_date_to_timestamp(date))
            with open(self.db.filename, 'r', encoding='utf-8') as f:
                db_data = json.load(f)
            return list(map(Entry.from_dict, filter(lambda d: d['timestamp'] < timestamp, db_data['data'])))

    def __convert_date_to_timestamp(*dates: Iterable[Union[datetime, int]]) -> Generator[int, None, None]:
        for date in dates:
            if type(date) is not int:
                date = date.timestamp()
            yield date



from .decorators.recorder import recorder as tmt_recorder
from .utils.manager import TmtManager
from .history.utils import save as tmt_save
from .configs.parser import Configs
from .info import __version__


__all__ = ['decorators', 'utils', 'history', 'info', 'configs']


from prompt_toolkit import Application
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.layout.containers import Window, HSplit, WindowAlign
from prompt_toolkit.layout.controls import FormattedTextControl, DummyControl
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.key_binding.bindings.focus import focus_next, focus_previous
from prompt_toolkit.widgets import Frame
from tmt.interface.tui.search import SearchLayout, SearchBy
from tmt.interface.tui.entry_display import EntryDisplay
from tmt.interface.tui.base import BaseApp, FocusableText
from tmt.storage.schema import Entry
from typing import Optional
from functools import partial

kb = KeyBindings()

@kb.add('c-q')
def exit_(event):
    event.app.exit()


class MainApp(BaseApp):
    # TODO serve a proper "window_too_small" window
    def __init__(self, config_path: Optional[str] = None):
        self.search_layout = SearchLayout(back_handler=self.main_window, entry_select_handler=self.entry_display, config_path=config_path)
        self.main_container = HSplit([
            Frame(
                title='That Metric Timeline (TMT)', 
                body=HSplit([
                    Window(content=FormattedTextControl(text='This is the Terminal User Interface (TUI) of That Metric Timeline library. You can look for recorded experiments and see their details\nChoose an option from below:'), height=3),
                    HSplit([
                        FocusableText('Search experiment by name', align=WindowAlign.CENTER, ext_height=False, handler=self.search_by_name),
                        FocusableText('Search experiment by ID', align=WindowAlign.CENTER, ext_height=False, handler=self.search_by_id),
                        FocusableText('Search experiment by date', align=WindowAlign.CENTER, ext_height=False, handler=self.search_by_date)], padding=1),
                    Window(content=DummyControl(), height=6)
                    ])),
            self._bottom_toolbar()
        ], padding=3)
        self.layout = Layout(self.main_container)
        kb.add("tab")(focus_next)
        kb.add("down")(focus_next)
        kb.add("s-tab")(focus_previous)
        kb.add("up")(focus_previous)

        self.search_by = None
        self.app = Application(full_screen=True, layout=self.layout, key_bindings=kb)

    def main_window(self):
        self.app.layout = self.layout

    def search_by_name(self, clear=True):
        self.search_by = SearchBy.NAME
        layout = self.search_layout.search_by_name_layout(clear)
        self.app.layout = Layout(layout, focused_element=layout)
    
    def search_by_id(self, clear=True):
        self.search_by = SearchBy.ID
        layout = self.search_layout.search_by_id_layout(clear)
        self.app.layout = Layout(layout, focused_element=layout)

    def search_by_date(self, clear=True):
        pass

    def entry_display(self, entry: Entry):
        if self.search_by is SearchBy.NAME:
            back_handler = partial(self.search_by_name, False)
        elif self.search_by is SearchBy.ID:
            back_handler = partial(self.search_by_id, False)
        elif self.search_by is SearchBy.DATE:
            back_handler = partial(self.search_by_date, False)
        else:
            back_handler = self.main_window
        layout = EntryDisplay(entry, back_handler=back_handler).get_layout()
        self.app.layout = Layout(layout, focused_element=layout)

    def run(self):
        self.app.run()


if __name__ == '__main__':
    import sys
    MainApp(sys.argv[1]).run()


from typing import Callable, Dict, Optional
from tmt.history.context import ContextManager, context_manager
from tmt.storage.json_db import DbManager
from tmt.storage.schema import Metric
from datetime import datetime


def recorder(name: str, config_path: Optional[str] = None, save_on_exception=False):
    """
    One of the main `tmt` functions. This is a decorator which can be used to keep track of 
    experiments, saving metrics, results and taking a snapshot of the code at this moment in time.

    .. note::
        For simplicity, `tmt` exposes this function from its root, with the `tmt_recorder` name. This is the 
        recommended way of importing this function, i.e. `from tmt import tmt_recorder`.

    **Usage**:

    .. code-block:: python

        from tmt import tmt_recorder 

        @tmt_recorder(name="some_experiment")
        def train_and_predict(x_tr, y_tr, x_te, y_te):
            lr = LogisticRegression()
            lr.fit(x_tr, y_tr)
            preds = lr.predict(x_te)
            return {'f1': f1_score(y_te, preds), 'accuracy': accuracy_score(y_te, preds)} 

    :param name: name used to save this experiment in the database.
    :type name: str
    :param config_path: if you want to use a custom configuration file, specify the path. Defaults to None. See :doc:`configuration` and :py:class:`tmt.configs.parser.Configs`
    :type config_path: Optional[str], optional
    :param save_on_exception: save everything (snapshot, metrics etc.) even if an exception happens. Defaults to False.
    :type save_on_exception: bool, optional
    """
    def inner(func: Callable[..., Optional[Dict[str, float]]]):
        def wrapper(*args, **kwargs) -> Optional[Dict[str, float]]:
            cm = ContextManager(name, config_path)
            context_manager.set(cm)
            db_manager = DbManager(cm.config.json_db_path)
            try:
                metrics = func(*args, **kwargs)
                if metrics:
                    for k, v in metrics.items():
                        cm.entry.metrics.append(Metric(cm.entry.id, k, v))
                cm.snap_manager.make_snapshot()
                cm.entry.date_saved = int(datetime.now().timestamp())
                db_manager.add_new_entries([cm.entry])
                return metrics
            except Exception as e:
                if not save_on_exception:
                    raise e
                cm.snap_manager.make_snapshot()
                cm.entry.date_saved = int(datetime.now().timestamp())
                db_manager.add_new_entries([cm.entry])
            
        return wrapper
    return inner



from __future__ import annotations
import os
import json
from tmt.history.snapshot import SnapshotManager
from tmt.storage.schema import BaseJsonDataclass
from dataclasses import dataclass

CONFIG_PATH = '.tmt/config.json'

@dataclass
class Configs(BaseJsonDataclass):
    """
    This dataclass holds the `tmt` configuration options. Users should not mind about this class, since this is 
    used internally by other `tmt` functions such as :py:func:`tmt.decorators.recorder.recorder` or :py:class:`tmt.utils.manager.TmtManager`. When instantiated with :py:func:`tmt.configs.parser.Configs.from_default_path_or_default_config()`, it will look
    for a json configuration file in the default path `.tmt/config.json`.
    If none is found, it will create `tmt` default configuration. See :doc:`Configuration <configuration>`.
    """
    tmt_dir: str
    snapshot_source: str
    snapshot_target: str
    last_snapshot_link: str
    gitignore_path: str
    json_db_path: str 
    results_path: str 

    @classmethod 
    def from_dict(cls, d):
        config = super().from_dict(d)
        config.snapshot_target = os.path.join(config.tmt_dir, config.snapshot_target)
        config.last_snapshot_link = os.path.join(config.tmt_dir, config.last_snapshot_link)
        config.json_db_path = os.path.join(config.tmt_dir, config.json_db_path)
        config.results_path = os.path.join(config.tmt_dir, config.results_path)
        return config

    @staticmethod
    def from_config(path: str = CONFIG_PATH) -> Configs:
        with open(path, 'r') as f:
            return Configs.from_dict(json.load(f))

    @staticmethod
    def from_default_path_or_default_config() -> Configs:
        if os.path.exists(CONFIG_PATH):
            with open(CONFIG_PATH, 'r') as f:
                return Configs.from_dict(json.load(f))
        return Configs.default_config()

    @staticmethod
    def default_config() -> Configs:
        return Configs(
            tmt_dir='.tmt',
            snapshot_source=os.getcwd(),
            snapshot_target='.tmt/snapshots',
            last_snapshot_link='.tmt/snapshots/last',
            gitignore_path=os.path.join(os.getcwd(), '.gitignore'),
            json_db_path=".tmt/tmt_db.json",
            results_path=".tmt/results"
        )

    def init_snapshot_manager(self, id: str):
        return SnapshotManager(
            id=id,
            tmt_dir=self.tmt_dir,
            snapshot_source=self.snapshot_source,
            snapshot_target=self.snapshot_target,
            last_snapshot_link=self.last_snapshot_link,
            ignore_path=self.gitignore_path
        )



from filecmp import dircmp
from dataclasses import dataclass
from functools import cached_property
from typing import Set, Optional
import itertools
import pathspec
import os
import shutil


@dataclass
class SnapshotManager:
    id: str
    tmt_dir: str
    snapshot_target: str
    snapshot_source: str
    last_snapshot_link: str
    ignore_path: Optional[str] = None

    @cached_property
    def gitignored_files(self) -> Set[str]: 
        ignore_tmt = {self.tmt_dir}
        if not self.ignore_path or not os.path.exists(self.ignore_path):
            return ignore_tmt
        with open(self.ignore_path, 'r') as f:
            spec = pathspec.PathSpec.from_lines('gitwildmatch', f)
        ignore_tmt.update(spec.match_tree(self.snapshot_source))
        return ignore_tmt
    
    @cached_property
    def snapshot_dest(self) -> str:
        return os.path.join(os.path.abspath(self.snapshot_target), self.id)

    def make_snapshot(self):
        if not os.path.exists(self.last_snapshot_link):
            self.copy_files()
            return
        last_snapshot_path = os.readlink(self.last_snapshot_link)
        cmp = dircmp(self.snapshot_source, last_snapshot_path)
        os.makedirs(self.snapshot_dest, exist_ok=True)
        for push_file in itertools.chain(self.get_diff_files(cmp, ''), self.get_new_files(cmp, '')):
            os.makedirs(os.path.join(self.snapshot_dest, os.path.dirname(push_file)), exist_ok=True)
            if os.path.isfile(os.path.join(self.snapshot_source, push_file)):
                shutil.copy(os.path.join(self.snapshot_source, push_file), os.path.join(self.snapshot_dest, push_file))
        for eq_file in self.get_equal_files(cmp, ''):
            os.makedirs(os.path.join(self.snapshot_dest, os.path.dirname(eq_file)), exist_ok=True)
            if os.path.isfile(os.path.join(self.snapshot_source, eq_file)):
                os.link(os.path.join(last_snapshot_path, eq_file), os.path.join(self.snapshot_dest, eq_file))
        self.create_symlink()

    def copy_files(self):
        shutil.copytree(self.snapshot_source, self.snapshot_dest, ignore=shutil.ignore_patterns(*self.gitignored_files))
        self.create_symlink()

    def create_symlink(self):
        if os.path.exists(self.last_snapshot_link):
            os.remove(self.last_snapshot_link)
        os.symlink(self.snapshot_dest, self.last_snapshot_link)

    @staticmethod
    def get_diff_files(cmp: dircmp, directory: str):
        for n in cmp.diff_files:
            yield os.path.join(directory, n)
        for d, sub_cmp in cmp.subdirs.items():
            yield from SnapshotManager.get_diff_files(sub_cmp, os.path.join(directory, d))

    @staticmethod
    def get_new_files(cmp: dircmp, directory: str):
        for n in cmp.left_only:
            yield os.path.join(directory, n)
        for d, sub_cmp in cmp.subdirs.items():
            yield from SnapshotManager.get_new_files(sub_cmp, os.path.join(directory, d))

    @staticmethod
    def get_removed_files(cmp: dircmp, directory: str):
        for n in cmp.right_only:
            yield os.path.join(directory, n)
        for d, sub_cmp in cmp.subdirs.items():
            yield from SnapshotManager.get_removed_files(sub_cmp, os.path.join(directory, d))

    @staticmethod
    def get_equal_files(cmp: dircmp, directory: str):
        for n in cmp.same_files:
            yield os.path.join(directory, n)
        for d, sub_cmp in cmp.subdirs.items():
            yield from SnapshotManager.get_equal_files(sub_cmp, os.path.join(directory, d))


from tmt.storage.json_db import DbManager
from tmt.storage.schema import Entry, Metric
from tmt.configs.parser import Configs
from typing import Optional, Generator, Any, Tuple, List
from tmt.exceptions import EntryNotFound
import pickle


class TmtManager:
    """
    This is a rather minimalistic helper class which you can use to load experiments, load pickled results, see metrics and so on.
    The recommended way of looking for an experiment is using `tmt` terminal user interface (see :ref:`TMT TUI <tmttui>`).
    Once you have a unique id or name, you can use one of the setter methods to set the :py:class:`tmt.storage.schema.Entry` (i.e., an experiment) to be 
    used with this instance. If you want low-level access to the json database (read-only operations), you can access the `db` member.

    **Usage**:

    .. code-block:: python

        from tmt import TmtManager

        # Let's say we know there is an experiment with id "example"


        # An Entry is a row in the database, i.e. an experiment that was tracked.
        manager = TmtManager()
        manager.set_entry_by_id('example') 

        # load the results and unpickle them
        for name, path in manager.results_paths():
            with open(path, 'rb') as f:
                # do stuff with your results. If it's a pickle it's 
                # more convenient to use the code block below this one
                res = pickle.load(f)

        # load the unpickled results
        for name, res in manager.load_results():
            # do something with your results.
            # if res is a numpy array...
            print(res.mean())


        for name, val in manager.get_metrics():
            print(f"{name}: {val}")

        # If you need to do other stuff, like searching for 
        # experiments between two datetimes and so on
        # you can access the `db` member like
        manager.db.get_entries_greater_than_date(date_or_timestamp)

    :param entry: this is an entry in the json database, i.e. a previously tracked and saved experiments.
    :type entry: Optional[Entry]
    :param config: this can be a path to a custom configuration json file. See :doc:`Configuration <configuration>`.
    :type config: Optional[str]
    """

    def __init__(self, entry: Optional[Entry] = None, config: Optional[str] = None):
        if config is None:
            self.config = Configs.from_default_path_or_default_config()
        else:
            self.config = Configs.from_config(config)
        self.db = DbManager(self.config.json_db_path, read_only=True)
        self.entry = entry

    def set_entry_by_name(self, name: str) -> Entry:
        if (e := self.db.get_entry_by_exact_name(name)):
            self.entry = e
            return self.entry
        raise EntryNotFound(f'No entry (experiment) found for name {name}')

    def set_entry_by_id(self, id: str) -> Entry:
        if (e := self.db.get_entry_by_id(id)):
            self.entry = e
            return self.entry
        raise EntryNotFound(f'No entry (experiment) found for id {id}')

    def entry_not_none(func):
        def inner(*args, **kwargs):
            assert args[0].entry is not None, "This instance has not an `Entry` associated with it. Use one of the several `set_` methods"
            return func(*args, **kwargs)
        return inner
    
    @entry_not_none
    def load_results(self) -> Generator[Tuple[str, Any], None, None]:
        """
        Creates a generator which yields a tuple with the name of the results (see :py:func:`tmt.history.utils.save`)
        and the unpickled object. See also :py:func:`tmt.utils.manager.TmtManager.results_paths` if you just want the paths.

        :yield: A tuple with (name, object) of each result stored with this experiment.
        :rtype: Generator[Tuple[str, Any], None, None]
        """
        for result in self.entry.results:
            with open(result.path, 'rb') as f:
                yield result.name, pickle.load(f)

    @entry_not_none
    def results_paths(self) -> Generator[Tuple[str, str], None, None]:
        """
        Creates a generator which yields a tuple with the name of the results (see :py:func:`tmt.history.utils.save`)
        and its path on disk. See also :py:func:`tmt.utils.manager.TmtManager.load_results` if you want unpickled objects.

        :yield: A tuple with (name, path) of each result stored with this experiment.
        :rtype: Generator[Tuple[str, str], None, None]
        """
        for result in self.entry.results:
            yield result.name, result.path

    @entry_not_none
    def get_metrics(self) -> List[Metric]:
        return self.entry.metrics

    @entry_not_none
    def code_snapshot_path(self) -> str:
        """
        Returns the path to the code snapshot backup saved with this experiment.

        :return: path to the snapshot saved with this experiment.
        :rtype: str
        """
        return self.entry.local_snapshot_path



from tmt.history.context import context_manager
from typing import Any


def save(obj: Any, name: str, allow_exist=False, extension='.pkl') -> str:
    """
    This function can be used to save any `pickable <https://docs.python.org/3/library/pickle.html#what-can-be-pickled-and-unpickled>`_. object you may want to save when running your experiments.
    It must be called by a function decorated with :py:func:`tmt.decorators.recorder.recorder` (or within code called from that function).

    .. note::
        For simplicity, `tmt` exposes this function from its root with the `tmt_save` name. You can import it with `from tmt import 
        tmt_save`

    What this function does is not too far from simply doing:

    .. code-block:: python

        with open(library_path + name, 'wb') as f:
            pickle.dump(obj, f)
    
    Using this function however, we make sure that `tmt` knows where all of your pickled objects are so that you can 
    later easily retrieve them.

    **Usage**:

    .. code-block:: python

        from tmt import tmt_recorder, tmt_save

        @tmt_recorder(name="some_experiment_with_data")
        def train_and_predict(...):
            ...
            preds = lr.predict(x_te)
            tmt_save(preds, name='lr_predictions')
            return {'f1': f1_score(y_te, preds), 'accuracy': accuracy_score(y_te, preds)}

    :param obj: the object you want to save, must be pickable.
    :type obj: Any
    :param name: the object will be saved with this name. Use a name that will help you recognize this object in the future.
    :type name: str
    :param allow_exist: if True, allows this object to be saved even if the name already exists. Defaults to False.
    :type allow_exist: bool, optional
    :param extension: the extension to save the file with, defaults to '.pkl'.
    :type extension: str, optional
    :raises ValueError: this is raised if the function is illegaly called outside a function decorated with :py:func:`tmt.decorators.recorder.recorder`
    :raises tmt.exceptions.DuplicatedNameError: if `name` already exists and `allow_exist` is False, this exception will be raised.
    :return: the path where the object was saved. `tmt` will take care of storing it, so you can ignore this return value.
    :rtype: str
    """
    context = context_manager.get()
    if context is None:
        raise ValueError("`save` function was called before initializing the context. Did you use a tmt decorator/class?")
    return context.save(obj, name, allow_exist, extension)